/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_FlatTree_h__
#define __surface_FlatTree_h__

namespace fe
{

/**************************************************************************//**
	@brief Triangular storage using a simple array

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT FlatTree: virtual public SpatialTreeI
{
	public:

	class Hit: public SpatialTreeI::Hit
	{
		public:
								Hit(void)
								{	reset(); }

		void					reset(void)
								{
									m_face= -1;
									m_triangleIndex= -1;
									m_primitiveIndex= -1;
									m_partitionIndex= -1;
									m_distance= -1.0;
									set(m_intersection);
									set(m_direction);
									set(m_barycenter);
									set(m_uv);
									set(m_du);
									set(m_dv);
									set(m_pVertex[0]);
									set(m_pVertex[1]);
									set(m_pVertex[2]);
									set(m_pNormal[0]);
									set(m_pNormal[1]);
									set(m_pNormal[2]);
								}

	virtual
	const	SpatialVector&		intersection(void) const
								{	return m_intersection; }
	virtual
	const	SpatialVector&		direction(void) const
								{	return m_direction; }
	virtual	Real				distance(void) const
								{	return m_distance; }
	virtual I32					face(void) const
								{	return m_face; }
	virtual I32					triangleIndex(void) const
								{	return m_triangleIndex; }
	virtual I32					primitiveIndex(void) const
								{	return m_primitiveIndex; }
	virtual I32					partitionIndex(void) const
								{	return m_partitionIndex; }
	virtual
	const	SpatialBary&		barycenter(void) const
								{	return m_barycenter; }
	virtual
	const	Vector2&			uv(void) const
								{	return m_uv; }
	virtual
	const	SpatialVector&		du(void) const
								{	return m_du; }
	virtual
	const	SpatialVector&		dv(void) const
								{	return m_dv; }
	virtual
	const	SpatialVector*		vertex(void) const
								{	return m_pVertex; }
	virtual
	const	SpatialVector*		normal(void) const
								{	return m_pNormal; }

			void				copy(const sp<Hit>& a_rspOther)
								{
									m_intersection=a_rspOther->intersection();
									m_direction=a_rspOther->direction();
									m_distance=a_rspOther->distance();
									m_face=a_rspOther->face();
									m_triangleIndex=a_rspOther->triangleIndex();
									m_primitiveIndex=
											a_rspOther->primitiveIndex();
									m_partitionIndex=
											a_rspOther->partitionIndex();
									m_barycenter=a_rspOther->barycenter();
									m_uv=a_rspOther->uv();
									m_du=a_rspOther->du();
									m_dv=a_rspOther->dv();
									m_pVertex[0]=a_rspOther->vertex()[0];
									m_pVertex[1]=a_rspOther->vertex()[1];
									m_pVertex[2]=a_rspOther->vertex()[2];
									m_pNormal[0]=a_rspOther->normal()[0];
									m_pNormal[1]=a_rspOther->normal()[1];
									m_pNormal[2]=a_rspOther->normal()[2];
								}

			SpatialVector		m_intersection;
			SpatialVector		m_direction;
			Real				m_distance;
			I32					m_face;
			I32					m_triangleIndex;
			I32					m_primitiveIndex;
			I32					m_partitionIndex;
			SpatialBary			m_barycenter;
			Vector2				m_uv;
			SpatialVector		m_du;
			SpatialVector		m_dv;
			SpatialVector		m_pVertex[3];
			SpatialVector		m_pNormal[3];
	};

				FlatTree(void);
virtual			~FlatTree(void)												{}

virtual	void	setRefinement(U32 a_refinement)								{}
virtual	void	populate(
						const Vector3i* a_pElementArray,
						const SpatialVector* a_pVertexArray,
						const SpatialVector* a_pNormalArray,
						const Vector2* a_pUVArray,
						const I32* a_pTriangleIndexArray,
						const I32* a_pPrimitiveIndexArray,
						const I32* a_pPartitionIndexArray,
						U32 a_primitives,U32 a_vertices,
						const SpatialVector& a_center,F32 a_radius);
virtual	Real	nearestPoint(const SpatialVector& a_origin,U32 a_hitLimit,
						const sp<SurfaceI::PartitionI>& a_rspPartition,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;
virtual	Real	rayImpact(const SpatialVector& a_origin,
						const SpatialVector& a_direction,Real a_maxDistance,
						U32 a_hitLimit,
						const sp<SurfaceI::PartitionI>& a_rspPartition,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;
virtual	void	draw(sp<DrawI>,const fe::Color*) const						{}

	protected:

		void	nearestAccumulate(I32 a_face,I32 a_triangleIndex,
						I32 a_primitiveIndex,I32 a_partitionIndex,
						const SpatialVector& a_origin,U32 a_hitLimit,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
						const SpatialVector& a_rCenter,Real a_radius,
						const SpatialVector& a_rVertex0,
						const SpatialVector& a_rVertex1,
						const SpatialVector& a_rVertex2,
						const SpatialVector& a_rNormal0,
						const SpatialVector& a_rNormal1,
						const SpatialVector& a_rNormal2) const;

static	void	addHit(U32 a_hitLimit,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
						CountedPool<Hit>& a_rHitPool,Real a_distance,
						const SpatialVector& a_rDirection,
						const SpatialVector& a_rIntersection,
						I32 a_face,I32 a_triangleIndex,
						I32 a_primitiveIndex,I32 a_partitionIndex,
						const SpatialBary& a_rBarycenter,
						const Vector2& a_rUV,
						const SpatialVector& a_rDu,
						const SpatialVector& a_rDv,
						const SpatialVector& a_rVertex0,
						const SpatialVector& a_rVertex1,
						const SpatialVector& a_rVertex2,
						const SpatialVector& a_rNormal0,
						const SpatialVector& a_rNormal1,
						const SpatialVector& a_rNormal2);

	struct Sort
	{
		bool operator()(const sp<SpatialTreeI::Hit>& rspHit1,
				const sp<SpatialTreeI::Hit>& rspHit2)
		{
			return (rspHit1->distance() < rspHit2->distance());
		}
	};

		void	cacheSpheres(void);

const	Vector3i*			m_pElementArray;
const	SpatialVector*		m_pVertexArray;
const	SpatialVector*		m_pNormalArray;
const	Vector2*			m_pUVArray;
const	I32*				m_pTriangleIndexArray;
const	I32*				m_pPrimitiveIndexArray;
const	I32*				m_pPartitionIndexArray;
		U32					m_primitives;
		U32					m_vertices;

		Vector4*			m_pSphereArray;
		U32					m_sphereCount; // != primitive count if clipped

mutable	CountedPool<Hit>	m_hitPool;
};

} /* namespace */

#endif /* __surface_FlatTree_h__ */
